/**
 *
 * SIROCCO
 * Copyright (C) 2011 France Telecom
 * Contact: sirocco@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  $Id$
 *
 */
package org.ow2.sirocco.cloudmanager.connector.util.jobmanager.impl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.Dictionary;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.naming.InitialContext;

import org.osgi.service.cm.ConfigurationException;
import org.osgi.service.cm.ManagedService;
import org.ow2.sirocco.cloudmanager.connector.util.jobmanager.api.IJobManager;
import org.ow2.sirocco.cloudmanager.model.cimi.CloudResource;
import org.ow2.sirocco.cloudmanager.model.cimi.Job;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.util.concurrent.ListenableFuture;

public class JobManager implements IJobManager, ManagedService {
    private static Logger logger = LoggerFactory.getLogger(JobManager.class);

    public static String JOB_WATCHER_PERIOD_PROP_NAME = "jobWatcherPeriodInSeconds";

    public static String JOB_RETENTION_TIME_PROP_NAME = "jobRetentionPeriodInSeconds";

    public static long DEFAULT_JOB_WATCHER_PERIOD_IN_SECONDS = 5 * 60;

    public static long DEFAULT_JOB_RETENTION_TIME_IN_SECONDS = 5 * 60;

    private static final String JMS_TOPIC_CONNECTION_FACTORY_NAME = "jms/TopicConnectionFactory";

    private static final String JMS_TOPIC_NAME = "jms/JobCompletion";

    private static class JobEntry {
        JobEntry(final Job job, final ListenableFuture<?> result) {
            this.job = job;
            this.result = result;

        }

        Job job;

        ListenableFuture<?> result;

        boolean sendNotif;
    };

    private long jobWatcherPeriodInSeconds = JobManager.DEFAULT_JOB_WATCHER_PERIOD_IN_SECONDS;

    private long jobRetentionPeriodInSeconds = JobManager.DEFAULT_JOB_RETENTION_TIME_IN_SECONDS;

    private final Map<String, JobEntry> jobs = new ConcurrentHashMap<String, JobEntry>();

    private final ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();

    private final ExecutorService jobCompletionExecutorService = Executors.newSingleThreadExecutor();

    private ScheduledFuture<?> jobWatcherHandle;

    private JobManager() {
    }

    @SuppressWarnings("rawtypes")
    public void updated(final Dictionary properties) throws ConfigurationException {
        if (properties != null) {
            String s = (String) properties.get(JobManager.JOB_WATCHER_PERIOD_PROP_NAME);
            if (s != null) {
                try {
                    this.jobWatcherPeriodInSeconds = Integer.parseInt(s);
                } catch (NumberFormatException ex) {
                    JobManager.logger.error("Illegal value for jobWatcherPeriodInSeconds property: " + s);
                }
            }
            s = (String) properties.get(JobManager.JOB_RETENTION_TIME_PROP_NAME);
            if (s != null) {
                try {
                    this.jobRetentionPeriodInSeconds = Integer.parseInt(s);
                } catch (NumberFormatException ex) {
                    JobManager.logger.error("Illegal value for jobRetentionPeriodInSeconds property: " + s);
                }
            }

        }
        JobManager.logger.info("JobManager ready, watcher period: " + this.jobWatcherPeriodInSeconds
            + "s, job retention time: " + this.jobRetentionPeriodInSeconds + " s");
    }

    public void start() {
        final Runnable jobWatcher = new Runnable() {
            @Override
            public void run() {
                JobManager.logger.debug("Job watcher woken up");
                Date now = new Date();
                for (JobEntry jobEntry : JobManager.this.jobs.values()) {
                    Job job = jobEntry.job;
                    if (job.getState() != Job.Status.RUNNING) {
                        if (TimeUnit.MILLISECONDS.toSeconds(now.getTime() - job.getTimeOfStatusChange().getTime()) > JobManager.this.jobRetentionPeriodInSeconds) {
                            JobManager.logger.info("Reaping job " + job.getProviderAssignedId());
                            JobManager.this.jobs.remove(job.getProviderAssignedId());
                        }
                    }
                }
            }
        };
        this.jobWatcherHandle = this.scheduler.scheduleAtFixedRate(jobWatcher, 0, this.jobWatcherPeriodInSeconds,
            TimeUnit.SECONDS);
    }

    public void shutdown() {
        this.jobWatcherHandle.cancel(true);
        this.scheduler.shutdown();
        System.out.println("JobManager shutdowned");
    }

    public static JobManager newJobManager() {
        JobManager jobManager = new JobManager();
        return jobManager;
    }

    public synchronized Job newJob(final CloudResource targetEntity, final CloudResource affectedEntity, final String action,
        final ListenableFuture<?> result) {
        String jobId = UUID.randomUUID().toString();
        final Job job = new Job();
        job.setProviderAssignedId(jobId);
        job.setTargetResource(targetEntity);
        job.setAction(action);
        job.setIsCancellable(false);
        job.setCreated(new Date());
        job.setState(Job.Status.RUNNING);
        List<CloudResource> affectedEntities = new ArrayList<CloudResource>();
        job.setAffectedResources(affectedEntities);
        if (affectedEntity != null) {
            affectedEntities.add(affectedEntity);
        }
        final JobEntry jobEntry = new JobEntry(job, result);
        this.jobs.put(jobId, jobEntry);
        jobEntry.result.addListener(new Runnable() {
            @Override
            public void run() {
                synchronized (jobEntry) {
                    if (jobEntry.result.isCancelled()) {
                        jobEntry.job.setState(Job.Status.CANCELLED);
                        jobEntry.job.setStatusMessage("cancelled");
                    } else {
                        boolean interrupted = false;
                        try {
                            while (true) {
                                try {
                                    jobEntry.result.get();
                                    jobEntry.job.setState(Job.Status.SUCCESS);
                                    break;
                                } catch (InterruptedException ex) {
                                    interrupted = true;
                                    // retry until not interrupted
                                } catch (ExecutionException ex) {
                                    JobManager.logger.info("Job failed", ex.getCause());
                                    jobEntry.job.setStatusMessage(ex.getCause().getMessage());
                                    jobEntry.job.setState(Job.Status.FAILED);
                                    break;
                                } catch (CancellationException ex) {
                                    jobEntry.job.setState(Job.Status.CANCELLED);
                                    jobEntry.job.setStatusMessage("cancelled");
                                    break;
                                }
                            }
                        } finally {
                            if (interrupted) {
                                Thread.currentThread().interrupt();
                            }
                        }
                    }
                    jobEntry.job.setTimeOfStatusChange(new Date());
                    if (jobEntry.sendNotif) {
                        JobManager.this.emitJobCompletionEvent(jobEntry.job);
                    }
                }
            }
        }, this.jobCompletionExecutorService);
        Job jobByValue = new Job();
        synchronized (jobEntry) {
            jobByValue.setAction(job.getAction());
            jobByValue.setProviderAssignedId(job.getProviderAssignedId());
            jobByValue.setTargetResource(job.getTargetResource());
            jobByValue.setAction(job.getAction());
            jobByValue.setIsCancellable(job.getIsCancellable());
            jobByValue.setState(job.getState());
            jobByValue.setStatusMessage(job.getStatusMessage());
            jobByValue.setAffectedResources(job.getAffectedResources());
        }

        return jobByValue;
    }

    @Override
    public void setNotificationOnJobCompletion(final String jobId) throws Exception {
        final JobEntry jobEntry = this.jobs.get(jobId);
        if (jobEntry == null) {
            throw new Exception("Invalid jobId: " + jobId);
        }
        synchronized (jobEntry) {
            if (jobEntry.job.getState() != Job.Status.RUNNING) {
                this.emitJobCompletionEvent(jobEntry.job);
            } else {
                jobEntry.sendNotif = true;
            }
        }
    }

    public Job getJobById(final String id) {
        JobEntry jobEntry = this.jobs.get(id);
        if (jobEntry == null) {
            return null;
        }
        return jobEntry.job;
    }

    public List<Job> getAllJobs() {
        ArrayList<Job> result = new ArrayList<Job>();
        for (JobEntry jobEntry : this.jobs.values()) {
            result.add(jobEntry.job);
        }
        return result;
    }

    private <T> void emitJobCompletionEvent(final Job job) {
        try {
            JobManager.logger.info("SENDING EVENT JobCompletion action=" + job.getAction() + " target="
                + job.getTargetResource().getClass() + " id=" + job.getTargetResource().getProviderAssignedId());
            this.emitMessage(job);
        } catch (Exception ex) {
            JobManager.logger.error("Failed to emit message", ex);
        }
    }

    private void emitMessage(final Serializable payload) throws Exception {
        InitialContext ctx = new InitialContext();
        TopicConnectionFactory topicConnectionFactory = (TopicConnectionFactory) ctx
            .lookup(JobManager.JMS_TOPIC_CONNECTION_FACTORY_NAME);
        TopicConnection connection = topicConnectionFactory.createTopicConnection();
        TopicSession session = connection.createTopicSession(false, Session.AUTO_ACKNOWLEDGE);
        Topic cloudAdminTopic = (Topic) ctx.lookup(JobManager.JMS_TOPIC_NAME);
        TopicPublisher topicPublisher = session.createPublisher(cloudAdminTopic);
        ObjectMessage message = session.createObjectMessage();
        message.setObject(payload);
        topicPublisher.publish(message);
        topicPublisher.close();
        session.close();
        connection.close();
    }

    public static void rawEmitDelayedQueueMessage(final Serializable payload, final long delayMilli, final String queueFactory,
        final String queueName) throws Exception {

        new Thread() {
            @Override
            public void run() {

                /*
                 * Thread.sleep(delayMilli); InitialContext ctx = new
                 * InitialContext(); QueueConnectionFactory qcf =
                 * (QueueConnectionFactory) ctx .lookup(queueFactory);
                 * QueueConnection queueCon = qcf.createQueueConnection();
                 * QueueSession queueSession = queueCon. createQueueSession(
                 * false, Session.AUTO_ACKNOWLEDGE ); Queue queue = (Queue)
                 * ctx.lookup(queueName); QueueSender sender =
                 * queueSession.createSender( queue ); Message msg =
                 * queueSession.createObjectMessage(payload);
                 * //msg.setLongProperty("scheduleDate",
                 * System.currentTimeMillis() + delayMilli); sender.send( msg );
                 * JobManager.logger.info("EMITTED EVENT for payload " +
                 * payload.toString());
                 */
            }
        }.start();

    }

}
